use std::collections::HashMap;
use std::fs::{File, OpenOptions};
use std::io::{BufReader, Write};
use std::path::PathBuf;
use anyhow::anyhow;
use chrono::{DateTime, NaiveDate, Utc};
use serde::{Deserialize, Serialize};
use crate::config::Config;
use crate::models;
use crate::models::{MailAddress, StrMessage};

#[derive(Serialize, Deserialize, Debug, Clone)]
pub struct SerializableThread{
    pub id: u32,
    pub messages: Vec<String>
}

#[derive(Serialize, Deserialize, Debug, Clone)]
pub struct SerializableMessage {
    pub name: String,
    pub from: String,
    pub to: Vec<MailAddress>,
    pub cc: Vec<MailAddress>,
    pub hash: String,
    pub subject: String,
    pub date: DateTime<Utc>,
    pub uid: u32,
    pub list: String,
    pub id: String,
    pub in_reply_to: String,
}

impl PartialEq for SerializableMessage {
    fn eq(&self, other: &Self) -> bool {
        self.name == other.name &&
            self.from == other.from &&
            self.hash == other.hash &&
            self.subject == other.subject &&
            self.date == other.date &&
            self.uid == other.uid &&
            self.list == other.list
        // TODO add more properties
    }
}

impl Eq for SerializableMessage {}

impl From<models::StrMessage> for SerializableMessage {
    fn from(msg: models::StrMessage) -> Self {
        let naive_date = NaiveDate::from_ymd_opt(
            msg.date.year.clone() as i32,
            msg.date.month.clone() as u32,
            msg.date.day.clone() as u32
        ).unwrap();
        let naive_datetime = naive_date.and_hms_opt(
            msg.date.hour.clone() as u32,
            msg.date.minute.clone() as u32,
            msg.date.second.clone() as u32
        ).unwrap();
        
        SerializableMessage {
            name: msg.from.name.clone().unwrap_or_else(|| "Undefined Name".to_string()),
            from: msg.from.address.clone(),
            to: msg.to.clone(),
            cc: msg.cc.clone(),
            hash: msg.hash.clone(),
            subject: msg.subject.clone(),
            date: DateTime::from_naive_utc_and_offset(naive_datetime, Utc),
            uid: msg.uid.clone(),
            list: msg.list.clone(),
            id: msg.id.clone(),
            in_reply_to: msg.in_reply_to.unwrap_or_default()
        }
    }
}

pub struct Indexes{

}

impl Indexes {
    pub fn new() -> Self{
        Indexes{}
    }

    pub fn add(message: StrMessage) -> anyhow::Result<()>{
        let messages = Indexes::update_message_indexes(message.clone()).unwrap_or_default();
        if Config::global().out_dir.clone().join("threads.json").exists() {
            match Indexes::update_thread_indexes(message.clone(), messages.clone()){
                Ok(new_thread_flag) => {
                    let mut lists: Vec<String> = vec!["Sent".to_string()];
                    if message.list != "Sent"{
                        lists.push(message.list.clone())
                    }
                    let _ = Self::persist_indexes(lists); // tODO do not override all indexes
                }
                Err(_) => {}
            };
        }
        
        Ok(())
    }
    
    fn update_message_indexes(message: StrMessage) -> anyhow::Result<Vec<SerializableMessage>>{
        let path = Config::global().out_dir.clone().join("messages.json");
        let mut messages: Vec<SerializableMessage> = Self::get_messages()?;
        
        let s_message = SerializableMessage::from(message);
        if !messages.contains(&s_message){
            messages.push(s_message);
            let _ = Indexes::write_to_json(serde_json::to_string_pretty(&messages)?, path);
        }
        
        Ok(messages)
    }
    
    pub fn update_thread_indexes(message: StrMessage, messages: Vec<SerializableMessage>) -> anyhow::Result<bool>{
        let mut new_thread_flag = false;
        let path = Config::global().out_dir.clone().join("threads.json");
        
        let mut threads: Vec<SerializableThread> = Self::get_threads()?;
        
        match message.in_reply_to {
            None => {
                threads.push(SerializableThread{
                    id: (threads.len() + 1) as u32,  
                    messages: vec![message.id],
                });
                new_thread_flag = true;
            }
            Some(in_reply_to) => {
                match Self::find_by_id(&messages, in_reply_to) {
                    None => {
                        threads.push(SerializableThread{
                            id: (threads.len() + 1) as u32,
                            messages: vec![message.id],
                        })
                    }
                    Some(msg) => {
                        for thread in threads.iter_mut() {
                            if thread.messages.contains(&msg.id) {
                                thread.messages.push(msg.id.clone());
                            }
                        }
                    }
                }
            }
        }
        
        let _ = Indexes::write_to_json(serde_json::to_string_pretty(&threads)?, path);
        Ok(new_thread_flag)
    }
    
    pub fn persist_indexes(lists: Vec<String>) -> anyhow::Result<()>{        
        let mut messages: Vec<SerializableMessage> = Self::get_messages()?;
        let mut threads: Vec<SerializableThread> = Self::get_threads()?;

        for list in lists {
            let date_path = Config::global().out_dir.clone().join(list.clone()).join("date.json");
            let subject_path = Config::global().out_dir.clone().join(list.clone()).join("subject.json");
            let from_path = Config::global().out_dir.clone().join(list.clone()).join("from.json");
            let to_path = Config::global().out_dir.clone().join(list.clone()).join("to.json");
            let cc_path = Config::global().out_dir.clone().join(list.clone()).join("cc.json");
            
            Self::write_to_json(serde_json::to_string_pretty(&Self::group_threads_by_subject(&mut threads, list.clone(), &mut messages)?)?, subject_path)?;
            Self::write_to_json(serde_json::to_string_pretty(&Self::group_threads_by_date(&mut threads, list.clone(), &mut messages)?)?, date_path)?;
            Self::write_to_json(serde_json::to_string_pretty(&Self::group_threads_by_cc(&mut threads, list.clone(), &mut messages)?)?, cc_path)?;
            if list != "Sent"{
                Self::write_to_json(serde_json::to_string_pretty(&Self::group_threads_by_sender(&mut threads, list.clone(), &mut messages)?)?, from_path)?;
            }else{
                Self::write_to_json(serde_json::to_string_pretty(&Self::group_threads_by_receiver(&mut threads, list.clone(), &mut messages)?)?, to_path)?;
            }
        }
        
        Ok(())
    }

    pub fn group_threads_by_subject(mut threads: &mut Vec<SerializableThread>, list: String, messages: &mut Vec<SerializableMessage>) -> anyhow::Result<Vec<(String, Vec<u32>)>>{
        let threads_for_list = Self::get_threads_for_list(list.clone(), threads, messages);
        let mut map: HashMap<String, Vec<u32>> = HashMap::new();

        for thread in threads_for_list {
            map.entry(Self::get_last_from_list_from_thread(&thread, list.clone(), messages).unwrap().subject.trim().to_string())
                .or_insert_with(Vec::new)
                .push(thread.id);
        }

        let mut sorted_vec: Vec<(String, Vec<u32>)> = map.into_iter().collect();
        sorted_vec.sort_by(|a, b| a.0.cmp(&b.0));
        
        Ok(sorted_vec)
    }
    
    pub fn group_threads_by_date(mut threads: &mut Vec<SerializableThread>, list: String, messages: &mut Vec<SerializableMessage>) -> anyhow::Result<Vec<(NaiveDate, Vec<u32>)>> {
        let threads_for_list = Self::get_threads_for_list(list.clone(), threads, messages);
        let mut map: HashMap<NaiveDate, Vec<(DateTime<Utc>, u32)>> = HashMap::new();

        for thread in threads_for_list {
            if let Some(last_message) = Self::get_last_from_list_from_thread(&thread, list.clone(), messages) {
                let date_only = last_message.date.date_naive();
                map.entry(date_only)
                    .or_insert_with(Vec::new)
                    .push((last_message.date, thread.id));
            }
        }

        let sorted_map: HashMap<NaiveDate, Vec<u32>> = map.into_iter()
            .map(|(date, mut pairs)| {
                pairs.sort_by_key(|&(timestamp, _)| timestamp);
                let sorted_ids = pairs.into_iter().map(|(_, id)| id).collect();
                (date, sorted_ids)
            })
            .collect();
        
        let mut sorted_vec: Vec<(NaiveDate, Vec<u32>)> = sorted_map.into_iter().collect();
        sorted_vec.sort_by(|a, b| a.0.cmp(&b.0));
        
        Ok(sorted_vec)
    }

    pub fn group_threads_by_sender(mut threads: &mut Vec<SerializableThread>, list: String, messages: &mut Vec<SerializableMessage>) -> anyhow::Result<Vec<(String, Vec<u32>)>>{
        let mut threads_for_list = Self::get_threads_for_list(list.clone(), threads, messages);
        let mut map: HashMap<String, Vec<u32>> = HashMap::new();

        for thread in threads_for_list {
            map.entry(Self::get_last_from_list_from_thread(&thread, list.clone(), messages).unwrap().from)
                .or_insert_with(Vec::new)
                .push(thread.id);
        }

        let mut sorted_vec: Vec<(String, Vec<u32>)> = map.into_iter().collect();
        sorted_vec.sort_by(|a, b| a.0.cmp(&b.0));
        
        Ok(sorted_vec)
    }

    pub fn group_threads_by_receiver(mut threads: &mut Vec<SerializableThread>, list: String, messages: &mut Vec<SerializableMessage>) -> anyhow::Result<Vec<(String, Vec<u32>)>>{
        let mut threads_for_list = Self::get_threads_for_list(list.clone(), threads, messages);
        let mut map: HashMap<String, Vec<u32>> = HashMap::new();

        for thread in threads_for_list {
            for to in Self::get_last_from_list_from_thread(&thread, list.clone(), messages).unwrap().to {
                map.entry(to.address)
                    .or_insert_with(Vec::new)
                    .push(thread.id);
            }
        }

        let mut sorted_vec: Vec<(String, Vec<u32>)> = map.into_iter().collect();
        sorted_vec.sort_by(|a, b| a.0.cmp(&b.0));
        
        Ok(sorted_vec)
    }

    pub fn group_threads_by_cc(mut threads: &mut Vec<SerializableThread>, list: String, messages: &mut Vec<SerializableMessage>) -> anyhow::Result<Vec<(String, Vec<u32>)>>{
        let mut threads_for_list = Self::get_threads_for_list(list.clone(), threads, messages);
        let mut map: HashMap<String, Vec<u32>> = HashMap::new();

        for thread in threads_for_list {
            for to in Self::get_last_from_list_from_thread(&thread, list.clone(), messages).unwrap().cc {
                map.entry(to.address)
                    .or_insert_with(Vec::new)
                    .push(thread.id);
            }
        }

        let mut sorted_vec: Vec<(String, Vec<u32>)> = map.into_iter().collect();
        sorted_vec.sort_by(|a, b| a.0.cmp(&b.0));
        
        Ok(sorted_vec)
    }
    
    fn get_threads_for_list(list: String, mut threads: &mut Vec<SerializableThread>, messages: &mut Vec<SerializableMessage>) -> Vec<SerializableThread>{
        let mut sorted_threads: Vec<SerializableThread>;
        sorted_threads = threads
            .into_iter()
            .filter_map(|thread| {
                Self::get_last_from_list_from_thread(&thread, list.clone(), messages).map(|first_message| thread.clone())
            })
            .collect::<Vec<_>>();
        sorted_threads
    }
    
    pub fn persist_threads() -> anyhow::Result<()>{
        let messages: Vec<SerializableMessage> = Self::get_messages()?;

        let mut uf = UnionFind::new();
        let mut uid_map: HashMap<String, u32> = HashMap::new();
        let mut message_map: HashMap<String, SerializableMessage> = HashMap::new();

        for message in &messages {
            uf.add(message.id.clone());
            if !message.in_reply_to.is_empty() {
                uf.add(message.in_reply_to.clone());
                uf.union(message.id.clone(), message.in_reply_to.clone());
            }
            uid_map.insert(message.id.clone(), message.uid);
            message_map.insert(message.id.clone(), message.clone());
        }

        let mut threads_map: HashMap<String, Vec<SerializableMessage>> = HashMap::new();
        for message in messages {
            let root = uf.find(&message.id);
            threads_map.entry(root).or_insert_with(Vec::new).push(message);
        }

        let mut threads: Vec<SerializableThread> = Vec::new();
        let mut id_counter = 1;

        for (_, mut thread_messages) in threads_map {
            // Sort messages based on the thread's sequence
            thread_messages.sort_by_key(|m| m.date.clone());

            // Construct the ordered list of uids for the thread
            let ordered_uids: Vec<String> = thread_messages.iter().map(|m| m.id.clone()).collect();

            threads.push(SerializableThread {
                id: id_counter,
                messages: ordered_uids,
            });
            id_counter += 1;
        }

        let thread_path = Config::global().out_dir.clone().join("threads.json");
        let _ = Indexes::write_to_json(serde_json::to_string_pretty(&threads)?, thread_path);
        
        Ok(())
    }
    
    fn write_to_json(content: String, path: PathBuf) -> anyhow::Result<()>{
        let mut file = OpenOptions::new()
            .write(true)
            .create(true)
            .truncate(true)
            .open(&path)?;
        file.write_all(content.as_bytes())?;
        
        Ok(())
    }

    fn find_by_id(messages: &Vec<SerializableMessage>, search_id: String) -> Option<SerializableMessage> {
        for message in messages {
            if message.id == search_id {
                return Some(message.clone());
            }
        }
        None
    }

    fn find_by_uid(messages: &Vec<SerializableMessage>, search_uid: u32) -> Option<SerializableMessage> {
        for message in messages {
            if message.uid == search_uid {
                return Some(message.clone());
            }
        }
        None
    }
    
    fn get_messages() -> anyhow::Result<Vec<SerializableMessage>>{
        let path = Config::global().out_dir.clone().join("messages.json");
        let messages: Vec<SerializableMessage> = if path.exists() {
            let file = File::open(&path)?;
            let reader = BufReader::new(file);
            serde_json::from_reader(reader)?
        } else {
            Vec::new()
        };
        
        Ok(messages)
    }
    
    fn get_threads() -> anyhow::Result<Vec<SerializableThread>>{
        let path = Config::global().out_dir.clone().join("threads.json");

        let mut threads: Vec<SerializableThread> = if path.exists() {
            let file = File::open(&path)?;
            let reader = BufReader::new(file);
            serde_json::from_reader(reader)?
        } else {
            Vec::new()
        };
        
        Ok(threads)
    }
    
    fn get_last_from_list_from_thread(thread: &SerializableThread, list: String, messages: &Vec<SerializableMessage>) -> Option<SerializableMessage>{
        let mut temp = thread.messages.clone();
        temp.reverse();
        for id in  temp{
            match Self::find_by_id(messages, id){
                None => {}
                Some(message) => {
                    if message.list == list { 
                        return Some(message)
                    }
                }
            }
        }
        
        None
    }

    pub fn delete_from_messages(message_id: String) -> anyhow::Result<SerializableMessage>{
        let path = Config::global().out_dir.clone().join("messages.json");
        let file = File::open(&path)?;
        let reader = BufReader::new(file);
        let mut messages: Vec<SerializableMessage> = serde_json::from_reader(reader)?;
        let message_for_deletion;
        if let Some(pos) = messages.iter().position(|message| message.id == message_id) {
            message_for_deletion = messages.remove(pos)
        }else{
            return Err(anyhow!("Cannot delete message from messages.json"))
        }
        Self::write_to_json(serde_json::to_string_pretty(&messages)?, path)?;
        

        Ok(message_for_deletion)
    }

    pub fn delete_from_threads(message: SerializableMessage) -> anyhow::Result<()>{
        let path = Config::global().out_dir.clone().join("threads.json");
        let file = File::open(&path)?;
        let reader = BufReader::new(file);
        let mut threads: Vec<SerializableThread> = serde_json::from_reader(reader)?;
        let mut i = 0;
        while i < threads.len() {
            if let Some(pos) = threads[i].messages.iter().position(|s| *s == message.id) {
                if threads[i].messages.len() == 1 {
                    threads.remove(i);
                } else {
                    threads[i].messages.remove(pos);
                }
                break;
            }
            i += 1;
        }
        
        Self::write_to_json(serde_json::to_string_pretty(&threads)?, path)?;
        Ok(())
    }
}


struct UnionFind {
    parent: HashMap<String, String>,
    rank: HashMap<String, u32>,
}

impl UnionFind {
    fn new() -> Self {
        Self {
            parent: HashMap::new(),
            rank: HashMap::new(),
        }
    }

    fn find(&mut self, item: &String) -> String {
        if self.parent.get(item) != Some(item) {
            let parent = self.parent[item].clone();
            let temp = self.find(&parent);
            self.parent.insert(item.clone(), temp);
        }
        self.parent[item].clone()
    }

    fn union(&mut self, item1: String, item2: String) {
        let root1 = self.find(&item1);
        let root2 = self.find(&item2);

        if root1 != root2 {
            let rank1 = *self.rank.get(&root1).unwrap_or(&0);
            let rank2 = *self.rank.get(&root2).unwrap_or(&0);

            if rank1 > rank2 {
                self.parent.insert(root2, root1);
            } else if rank1 < rank2 {
                self.parent.insert(root1, root2);
            } else {
                self.parent.insert(root2.clone(), root1.clone());
                self.rank.insert(root1, rank1 + 1);
            }
        }
    }

    fn add(&mut self, item: String) {
        self.parent.entry(item.clone()).or_insert(item.clone());
        self.rank.entry(item.clone()).or_insert(0);
    }
}

